/*
 * distromatic - tool for RPM based repositories
 *
 * Copyright (C) 2013 by Silvan Calarco <silvan.calarco@mambasoft.it>
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of version 2 of the GNU General Public License as published by the
 * Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY, to the extent permitted by law; without even the implied
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA.
 */

#include <sqlite3.h>
#include "config.h"
#include "headerlist.h"
#include "changelog.h"
#include "backend-sqlite3.h"

#define SQLITE_TABLE_files "id INTEGER PRIMARY KEY, "\
   "name STRING, firstchild INTEGER, next INTEGER, parent INTEGER, numproviders INTEGER"
/*     struct headerList **provider; */

#define SQLITE_TABLE_provided "id INTEGER PRIMARY KEY, "\
   "name STRING, flags INTEGER, numproviders INTEGER"
/*struct providedList {
     struct headerList **provider;
     int numbuildproviders;
     struct headerList **buildprovider;
     int numversions;
     char **version;
     int buildpriority;
     struct providedList *next;
};*/

#define SQLITE_TABLE_packagers "id INTEGER PRIMARY KEY, "\
   "name STRING, role INTEGER, changes_count INTEGER, packages_count INTEGER"
/*     char *alias[PACKAGER_MAXALIASES]; */

#define SQLITE_TABLE_changelog "id INTEGER PRIMARY KEY, "\
   "id_source INTEGER, time INTEGER, id_packager INTEGER, release STRING, text STRING"

#define SQLITE_TABLE_sources_source "id INTEGER PRIMARY KEY, id_source INTEGER, source STRING"
#define SQLITE_TABLE_sources_patch "id INTEGER PRIMARY KEY, id_source INTEGER, patch STRING"
//#define SQLITE_TABLE_sources_files_rel "id INTEGER PRIMARY KEY, id_source INTEGER, id_files INTEGER"
#define SQLITE_TABLE_sources_provided_rel "id INTEGER PRIMARY KEY, id_source INTEGER, id_provided INTEGER"

#define SQLITE_TABLE_sources "id INTEGER PRIMARY KEY, "\
   "name STRING, altrepository INTEGER, epoch INTEGER, version STRING, release STRING, summary STRING, id_packager INTEGER, "\
   "groupdescr STRING, description STRING, url STRING, license STRING, arch STRING, buildarchs STRING, " \
   "excludearch STRING, buildtime INTEGER, size INTEGER, id_changelog INTEGER"
/*   char **dirname;
     char **basename;
     int filenamecount;
     int requirecount;
     struct Require **require;
     struct providedList **required;
     int providecount;
     struct requireList *requirelist;
     struct headerSourceList *old;
     struct headerList *firstchild[ARCHS_MAX];
     struct warningList *firstwarning;
     struct headerSourceList *updatingparent;
};*/

#define SQLITE_TABLE_packages_files_rel "id INTEGER PRIMARY KEY, id_package INTEGER, id_source INTEGER"
#define SQLITE_TABLE_packages "id INTEGER PRIMARY KEY, "\
   "name STRING, altrepository INTEGER, epoch INTEGER, version STRING, release STRING, summary STRING, "\
   "groupdescr STRING, description STRING, size INTEGER, id_source INTEGER"
//   , url STRING, license STRING, arch STRING, buildarchs STRING, " \
//   "excludearch STRING, buildtime INTEGER, size INTEGER, id_changelog INTEGER"

/*struct headerList {
     char **obsoletename;
     uint_32 *obsoleteflags;
     char **obsoleteversion;
     int obsoletecount;
     char **providename;
     uint_32 *provideflags;
     char **provideversion;
     int providecount;
     int requirecount;
     struct Require **require;
     struct providedList **provided;
     int recursed;
     struct requireList *requirelist;
     struct headerList *next, *nextbrother;
     int obsoleted;
};*/

char sqlite3_query[PATH_MAX];
sqlite3_stmt *stmt;
long sqlite3_transaction_size;

int SQLite_init_table(sqlite3 *db, const char* table_name, const char* table_query) {

    char myquery[PATH_MAX];

    snprintf(myquery, PATH_MAX, "CREATE TABLE IF NOT EXISTS %s(%s)", table_name, table_query);

    if (sqlite3_exec(db, myquery, 0, 0, 0)) {
       fprintf(stderr, "ERROR: SQLite: (%s) %s\n", myquery, sqlite3_errmsg(db));
       return 2;
    }
    return 0;
}

int SQLite_begin_transaction(sqlite3 *db) {

    if (sqlite3_exec(db, "BEGIN TRANSACTION;", 0, 0, 0)) {
       fprintf(stderr, "ERROR: SQLite: (BEGIN TRANSACTION) %s\n", sqlite3_errmsg(db));
       return 1;
    }

    sqlite3_transaction_size = 0;
    return 0;
}

int SQLite_commit_transaction(sqlite3 *db) {
   if (sqlite3_exec(db, "COMMIT;", 0, 0, 0)) {
       fprintf(stderr, "ERROR: SQLite: (COMMIT) %s\n", sqlite3_errmsg(db));   
       return 1;
   }
   sqlite3_transaction_size = 0;
   return 0;
}

void SQLite_print_contents_subtree(sqlite3 *db,
                            struct fileTree* ft, 
			    struct configTag* ct, 
			    int arch) {

   while (ft) {
      snprintf(sqlite3_query, PATH_MAX, "INSERT INTO files_%s VALUES(%d,?,%d,%d,%d,%d);",
         ct->arch[arch],
         ft->id,
         (ft->firstchild?ft->firstchild->id:-1),
         (ft->next?ft->next->id:-1),
         (ft->parent?ft->parent->id:-1),
         ft->numproviders);

      sqlite3_prepare_v2(db, sqlite3_query, -1, &stmt, NULL);
      sqlite3_bind_text(stmt, 1, ft->name, -1, SQLITE_STATIC);

      if (sqlite3_step(stmt) != SQLITE_DONE) {
         fprintf(stderr, "ERROR: SQLite: (%s) %s\n", sqlite3_query, sqlite3_errmsg(db));
      }

/*      sqlite3_transaction_size += strlen(sqlite3_query);
      if (sqlite3_transaction_size >= 655360) {
  fprintf(stderr, "Commit: %ld\n", sqlite3_transaction_size);
         sqlite3_exec(db, "COMMIT;", 0, 0, 0);
         sqlite3_transaction_size = 0;
      }*/

      if (ft->firstchild) {
         SQLite_print_contents_subtree(db, ft->firstchild, ct, arch);
      }
      ft=ft->next;
   }
}

int generateSQLite_files(struct configTag* ct, sqlite3 *db) {

   int i;

   for (i = 0; i < ARCHS_MAX && ct->arch[i]; i++) {
      snprintf(sqlite3_query, PATH_MAX, "files_%s", ct->arch[i]);
      SQLite_init_table(db, sqlite3_query, SQLITE_TABLE_files);
      SQLite_begin_transaction(db);
      SQLite_print_contents_subtree(db, ct->filetree[i], ct, i);
      SQLite_commit_transaction(db);
   
/*      if (sqlite3_finalize(stmt)) {
         fprintf(stderr, "ERROR: SQLite: (%s) %s", sqlite3_query, sqlite3_errmsg(db));
      }*/
   }

   return 0;
}


int generateSQLite_provided(struct configTag* ct, sqlite3 *db) {

   int i;
   struct providedList* provided;

   for (i = 0; i < ARCHS_MAX && ct->arch[i]; i++) {
      snprintf(sqlite3_query, PATH_MAX, "provided_%s", ct->arch[i]);
      SQLite_init_table(db, sqlite3_query, SQLITE_TABLE_provided);
      SQLite_begin_transaction(db);

      provided = ct->providedlist_idx[i][0];
      while (provided) {
         snprintf(sqlite3_query, PATH_MAX, "INSERT INTO provided_%s VALUES(%d,?,%d,%d);",
            ct->arch[i],
            provided->id,
            provided->flags,
            provided->numproviders);

         sqlite3_prepare_v2(db, sqlite3_query, -1, &stmt, NULL);
         sqlite3_bind_text(stmt, 1, provided->name, -1, SQLITE_STATIC);

         if (sqlite3_step(stmt) != SQLITE_DONE) {
            fprintf(stderr, "ERROR: SQLite: (%s) %s\n", sqlite3_query, sqlite3_errmsg(db));
         }         
         provided = provided->next;
      }

      SQLite_commit_transaction(db);
   
/*      if (sqlite3_finalize(stmt)) {
         fprintf(stderr, "ERROR: SQLite: (%s) %s", sqlite3_query, sqlite3_errmsg(db));
      }*/
   }

   return 0;
}

int generateSQLite_packagers(sqlite3 *db) {

   struct Packager* packager = firstPackager();

   SQLite_init_table(db, "packagers", SQLITE_TABLE_packagers);
   SQLite_begin_transaction(db);
  
   while (packager) {
      snprintf(sqlite3_query, PATH_MAX, "INSERT INTO packagers VALUES(NULL,?,%d,%d,%d);",
         packager->role,
         packager->changes_count,
         packager->packages_count);

      if (sqlite3_prepare_v2(db, sqlite3_query, -1, &stmt, NULL)) {
         fprintf(stderr, "ERROR: SQLite: (%s) %s\n", sqlite3_query, sqlite3_errmsg(db));       
      }
      sqlite3_bind_text(stmt, 1, packager->name, -1, SQLITE_STATIC);
      if (sqlite3_step(stmt) != SQLITE_DONE) {
         fprintf(stderr, "ERROR: SQLite: (%s) %s\n", sqlite3_query, sqlite3_errmsg(db));
         return 3;
      }
      packager->id = sqlite3_last_insert_rowid(db);
      sqlite3_finalize(stmt);
      packager = packager->next;
   }
   SQLite_commit_transaction(db);
}

long generateSQLite_add_changelog(sqlite3 *db, struct changeLog* firstchangelog, long id_source) {

   struct changeLog* changelog = firstchangelog;
   long lastchangelogid = 0;

   while (changelog) {
      snprintf(sqlite3_query, PATH_MAX, "INSERT INTO changelog VALUES(NULL,%d,%d,%d,?,?);",
         id_source,
         changelog->time,
         changelog->pkg->id);

      if (sqlite3_prepare_v2(db, sqlite3_query, -1, &stmt, NULL)) {
         fprintf(stderr, "ERROR: SQLite: (%s) %s\n", sqlite3_query, sqlite3_errmsg(db));       
      }
      sqlite3_bind_text(stmt, 1, changelog->release, -1, SQLITE_STATIC);
      sqlite3_bind_text(stmt, 2, changelog->text, -1, SQLITE_STATIC);
      if (sqlite3_step(stmt) != SQLITE_DONE) {
         fprintf(stderr, "ERROR: SQLite: (%s) %s\n", sqlite3_query, sqlite3_errmsg(db));
         return 3;
      }
      changelog->id = sqlite3_last_insert_rowid(db);
      lastchangelogid = changelog->id;
      sqlite3_finalize(stmt);
      changelog = changelog->next;
   }
   return lastchangelogid;
}

int
generateSQLite_sources(struct configTag *ct, sqlite3 *db) {

    struct headerSourceList* currsource;
    int i;
    long nextchangelogid = 1;

    SQLite_init_table(db, "sources", SQLITE_TABLE_sources);
    SQLite_init_table(db, "sources_source", SQLITE_TABLE_sources_source);
    SQLite_init_table(db, "sources_patch", SQLITE_TABLE_sources_patch);
//    SQLite_init_table(db, "sources_files_rel", SQLITE_TABLE_sources_files_rel);
//    SQLite_init_table(db, "sources_provided_rel", SQLITE_TABLE_sources_provided_rel);
    SQLite_init_table(db, "changelog", SQLITE_TABLE_changelog);
    SQLite_begin_transaction(db);

    currsource = ct->headersourcelist;
    while (currsource != NULL) {

       snprintf(sqlite3_query, PATH_MAX, "INSERT INTO sources VALUES(%d,?,%d,%d,?,?,?,%d,?,?,?,?,?,?,?,%d,%d,%d);",
          currsource->id,
          currsource->altrepository,
          currsource->epoch,
          currsource->packager->id,
          currsource->buildtime,
          currsource->size,
          (currsource->changelog?nextchangelogid:0));
       if (sqlite3_prepare_v2(db, sqlite3_query, -1, &stmt, NULL)) {
          fprintf(stderr, "ERROR: SQLite: (%s) %s\n", sqlite3_query, sqlite3_errmsg(db));       
       }

       sqlite3_bind_text(stmt, 1, currsource->name, -1, SQLITE_STATIC);
       sqlite3_bind_text(stmt, 2, currsource->version, -1, SQLITE_STATIC);
       sqlite3_bind_text(stmt, 3, currsource->release, -1, SQLITE_STATIC);
       sqlite3_bind_text(stmt, 4, currsource->summary, -1, SQLITE_STATIC);
       sqlite3_bind_text(stmt, 5, currsource->group, -1, SQLITE_STATIC);
       sqlite3_bind_text(stmt, 6, currsource->description, -1, SQLITE_STATIC);
       sqlite3_bind_text(stmt, 7, currsource->url, -1, SQLITE_STATIC);
       sqlite3_bind_text(stmt, 8, currsource->license, -1, SQLITE_STATIC);
       sqlite3_bind_text(stmt, 9, currsource->arch, -1, SQLITE_STATIC);
       sqlite3_bind_text(stmt, 10, currsource->buildarchs, -1, SQLITE_STATIC);
       sqlite3_bind_text(stmt, 11, currsource->excludearch, -1, SQLITE_STATIC);

       if (sqlite3_step(stmt) != SQLITE_DONE) {
          fprintf(stderr, "ERROR: SQLite: (%s) %s\n", sqlite3_query, sqlite3_errmsg(db));
          return 3;
       }
       currsource->id = sqlite3_last_insert_rowid(db);
       sqlite3_finalize(stmt);

       /* source */
       if (currsource->source) {
          i=0;
          while (currsource->source[i]) {
             snprintf(sqlite3_query, PATH_MAX, "INSERT INTO sources_source VALUES(NULL,%d,?);",
                currsource->id);

             if (sqlite3_prepare_v2(db, sqlite3_query, -1, &stmt, NULL)) {
                fprintf(stderr, "ERROR: SQLite: (%s) %s\n", sqlite3_query, sqlite3_errmsg(db));       
             }
             sqlite3_bind_text(stmt, 1, currsource->source[i], -1, SQLITE_STATIC);
             if (sqlite3_step(stmt) != SQLITE_DONE) {
                fprintf(stderr, "ERROR: SQLite: (%s) %s\n", sqlite3_query, sqlite3_errmsg(db));
                return 3;
             }
             sqlite3_finalize(stmt);
             i++;
          }
       }

       /* patch */
       if (currsource->patch) {
          i=0;
          while (currsource->patch[i]) {
             snprintf(sqlite3_query, PATH_MAX, "INSERT INTO sources_patch VALUES(NULL,%d,?);",
                currsource->id);

             if (sqlite3_prepare_v2(db, sqlite3_query, -1, &stmt, NULL)) {
                fprintf(stderr, "ERROR: SQLite: (%s) %s\n", sqlite3_query, sqlite3_errmsg(db));       
             }
             sqlite3_bind_text(stmt, 1, currsource->patch[i], -1, SQLITE_STATIC);
             if (sqlite3_step(stmt) != SQLITE_DONE) {
                fprintf(stderr, "ERROR: SQLite: (%s) %s\n", sqlite3_query, sqlite3_errmsg(db));
                return 3;
             }
             sqlite3_finalize(stmt);
             i++;
          }
       }

       /* sources <-> provided relations (buildrequirements) */
/*       for (i=0; i<currsource->providecount; i++) {
          snprintf(sqlite3_query, PATH_MAX, "INSERT INTO sources_provided_rel VALUES(NULL,%d,%d);",
             currsource->id,
             currsource->required[i]->id);

          if (sqlite3_prepare_v2(db, sqlite3_query, -1, &stmt, NULL)) {
             fprintf(stderr, "ERROR: SQLite: (%s) %s\n", sqlite3_query, sqlite3_errmsg(db));       
          }
          if (sqlite3_step(stmt) != SQLITE_DONE) {
             fprintf(stderr, "ERROR: SQLite: (%s) %s\n", sqlite3_query, sqlite3_errmsg(db));
             return 3;
          }
          sqlite3_finalize(stmt);
       }*/

       if (currsource->changelog) {
          nextchangelogid = generateSQLite_add_changelog(db, currsource->changelog, currsource->id) + 1;
       }
       currsource = currsource->next;
    }

    SQLite_commit_transaction(db);
}

int
generateSQLite_packages(struct configTag *ct, sqlite3 *db, int arch) {

    struct headerList* currpackage;
    int i;

    snprintf(sqlite3_query, PATH_MAX, "packages_%s", ct->arch[i]);
    SQLite_init_table(db, sqlite3_query, SQLITE_TABLE_packages);
    snprintf(sqlite3_query, PATH_MAX, "packages_files_rel_%s", ct->arch[i]);
    SQLite_init_table(db, sqlite3_query, SQLITE_TABLE_packages_files_rel);
    SQLite_begin_transaction(db);

    currpackage = ct->headerlist[arch];
    while (currpackage) {

       snprintf(sqlite3_query, PATH_MAX, "INSERT INTO packages_%s VALUES(NULL,?,%d,%d,?,?,?,?,?,%d,%d);",
          ct->arch[arch],
          currpackage->altrepository,
          currpackage->epoch,
          currpackage->size,
          currpackage->sourceheader->id);
       if (sqlite3_prepare_v2(db, sqlite3_query, -1, &stmt, NULL)) {
          fprintf(stderr, "ERROR: SQLite: (%s) %s\n", sqlite3_query, sqlite3_errmsg(db));       
       }

       sqlite3_bind_text(stmt, 1, currpackage->name, -1, SQLITE_STATIC);
       sqlite3_bind_text(stmt, 2, currpackage->version, -1, SQLITE_STATIC);
       sqlite3_bind_text(stmt, 3, currpackage->release, -1, SQLITE_STATIC);
       sqlite3_bind_text(stmt, 4, currpackage->summary, -1, SQLITE_STATIC);
       sqlite3_bind_text(stmt, 5, currpackage->group, -1, SQLITE_STATIC);
       sqlite3_bind_text(stmt, 6, currpackage->description, -1, SQLITE_STATIC);

       if (sqlite3_step(stmt) != SQLITE_DONE) {
          fprintf(stderr, "ERROR: SQLite: (%s) %s\n", sqlite3_query, sqlite3_errmsg(db));
          return 3;
       }
       currpackage->id = sqlite3_last_insert_rowid(db);
       sqlite3_finalize(stmt);

       /* packages <-> files relations */
       for (i = 0; i < currpackage->filenamecount; i++) {
          snprintf(sqlite3_query, PATH_MAX, "INSERT INTO packages_files_rel_%s VALUES(NULL,%d,%d);",
             ct->arch[arch],
             currpackage->id,
             currpackage->file[i]->id);

          if (sqlite3_prepare_v2(db, sqlite3_query, -1, &stmt, NULL)) {
             fprintf(stderr, "ERROR: SQLite: (%s) %s\n", sqlite3_query, sqlite3_errmsg(db));       
          }
          if (sqlite3_step(stmt) != SQLITE_DONE) {
             fprintf(stderr, "ERROR: SQLite: (%s) %s\n", sqlite3_query, sqlite3_errmsg(db));
             return 3;
          }
          sqlite3_finalize(stmt);
       }
       currpackage = currpackage->next;
    }

    SQLite_commit_transaction(db);
}

int
generateSQLite(struct configTag *ct)
{
    sqlite3 *db = NULL;
    char dbname[PATH_MAX];
    int i;

    snprintf(dbname, PATH_MAX, "/tmp/%s.db", ct->tag);
    unlink(dbname);

    if (sqlite3_open_v2(dbname, &db, SQLITE_OPEN_READWRITE | SQLITE_OPEN_CREATE, NULL)) {
       if (db) sqlite3_close(db);
       return 1;
    }

    generateSQLite_packagers(db);
    generateSQLite_files(ct, db);
    generateSQLite_provided(ct, db);

    for (i = 0; i < ARCHS_MAX && ct->arch[i]; i++) {
       generateSQLite_packages(ct, db, i);
    }

    generateSQLite_sources(ct, db);

    sqlite3_close(db);
    return 0;
}
